#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json

from .twitch import Messages, Channel
from .subtitles import SubtitleWriter
from .settings import argparser, settings


def generate_config():
    with open('settings.json', 'w') as f:
        f.write(json.dumps(settings, indent=2))


def download(video):
    messages = Messages(video)
    writer = SubtitleWriter(messages)
    for comment in messages:
        writer.add(comment)
    writer.close()


def download_all(channel, min=0, max=None, count=None):
    videos = Channel(channel).videos()
    to_download = list()

    for video in videos:
        if (not max or video <= max) and video >= min:
            print(f'Downloading chat for {video}')
            download(video)

        if count and len(to_download) >= count:
            break

    for video in to_download:
        download(video)


def main():
    args = argparser.parse_args()

    if args.video or args.video_id:
        download(args.video or args.video_id)
    elif args.channel:
        download_all(args.channel, args.video_min,
                     args.video_max, args.video_count)
    elif args.generate_config:
        generate_config()
        print('New config had been written to ./settings.json '
              'and now will be used by default.')


if __name__ == "__main__":
    main()



import os
import sys
import json
import inspect
from argparse import ArgumentParser


SELF = inspect.getfile(inspect.currentframe())
ROOT = os.path.dirname(os.path.abspath(SELF))

settings_filename = 'settings.json'
home_settings_path = os.path.join(os.path.expanduser('~'), '.config/tcd', settings_filename)
example_settings_path = os.path.join(ROOT, f'example.{settings_filename}')

if os.path.isfile(settings_filename):
    settings_file_default = settings_filename
elif os.path.isfile(home_settings_path):
    settings_file_default = home_settings_path
else:
    settings_file_default = example_settings_path


def _pre_init_parser(help=False):
    parser = ArgumentParser(
        description='Download VOD chats from Twitch.', add_help=help)
    parser.add_argument(
        '-s', '--settings', metavar='settings.json',
        default=settings_file_default,
        help='Override path to settings.json.')
    return parser


args, unknown_args = _pre_init_parser().parse_known_args()
settings_file = args.settings

# Read settings from file
with open(settings_file, 'r') as settings_file:
    settings = json.load(settings_file)

# Check for outdated settings.json
if settings['version'].startswith('1'):
    print('Please update your settings.json ' +
          '(see example.settings.json for examples)')
    sys.exit(1)

if 'millisecond_separator' not in settings:
    settings['millisecond_separator'] = ','
if 'group_repeating_emotes' not in settings:
    settings['group_repeating_emotes'] = {'enabled': False,
                                          'threshold': 3,
                                          'collocations': 1,
                                          'format': '{emote} x{count}'}
if 'collocations' not in settings['group_repeating_emotes']:
    settings['group_repeating_emotes']['collocations'] = 1
if 'collocations_threshold' not in settings['group_repeating_emotes']:
    settings['group_repeating_emotes']['collocations_threshold'] = \
        settings['group_repeating_emotes']['collocations']
if 'video_types' not in settings:
    settings['video_types'] = 'archive'
if 'dynamic_duration' not in settings:
    settings['dynamic_duration'] = {'enabled': False, 'max': 5,
                                    'max_length': 100}
if 'max_width' not in settings:
    settings['max_width'] = -1
if 'badges' not in settings:
    settings['badges'] = {'enabled': False, 'max_count': 1, 'map': {}}

#
# Post-init settings overrides
#


def _post_init_parser(help=False):
    parser = _pre_init_parser(help)

    settings_group = parser.add_argument_group(
        'Settings', 'These options will override values from settings.json.')
    settings_group.add_argument(
        '--client-id', metavar='ID', type=str, default=settings['client_id'],
        help='Twitch API Client-ID headers.')
    settings_group.add_argument(
        '--cooldown', metavar='msec', type=int, default=settings['cooldown'],
        help='Delay (in milliseconds) between API calls.')

    progress = settings_group.add_mutually_exclusive_group(required=False)
    progress.add_argument(
        '--progress', action='store_true', dest='progress',
        default=settings['display_progress'],
        help='Display progress bar while downloading chat.')
    progress.add_argument(
        '--no-progress', action='store_false', dest='progress',
        help='Opposite of --progress.')

    settings_group.add_argument(
        '-f', '--formats', metavar='FORMAT', type=lambda s: s.split(','),
        default=settings['formats'],
        help='Comma-separated list of subtitle formats.')
    settings_group.add_argument(
        '-t', '--directory', metavar='PATH', type=str,
        default=settings['directory'],
        help='Target directory to save all generated subtitles.')
    settings_group.add_argument(
        '--filename-format', metavar='FORMAT', type=str,
        default=settings['filename_format'],
        help=('Python str.format for generating output file names. Available '
              'variables: {directory}, {video_id}, {format}, {user_name}, '
              '{title}, and {created_at}.'))
    settings_group.add_argument(
        '--max-width', metavar='chars', type=int,
        default=settings['max_width'],
        help=('Add line breaks to fit messages into specified width. '
              'Note: Implemented only for SSA/ASS subtitles.'))
    settings_group.add_argument(
        '--subtitle-duration', metavar='sec', type=int,
        default=settings['subtitle_duration'],
        help=('Time (in seconds) to display each comment on the screen. '
              'Will be ignored by some subtitle formats (irc).'))
    settings_group.add_argument(
        '--dynamic-duration', action='store_true', dest='dynamic',
        default=settings['dynamic_duration']['enabled'],
        help='Increase subtitle duration based on message length.')
    settings_group.add_argument(
        '--no-dynamic-duration', action='store_false', dest='dynamic',
        help='Opposite of --dynamic-duration.')
    settings_group.add_argument(
        '--dynamic-duration-max', metavar='sec', type=int,
        default=settings['dynamic_duration']['max'],
        help='Maximum duration of subtitle message.')
    settings_group.add_argument(
        '--dynamic-duration-max-length', metavar='chars', type=int,
        default=settings['dynamic_duration']['max_length'],
        help='Maximum length of subtitle message.')
    settings_group.add_argument(
        '--millisecond-separator', metavar='FORMAT', type=str,
        default=settings['millisecond_separator'],
        help='Separator between seconds and milliseconds in timestamps.')

    group = settings_group.add_mutually_exclusive_group(required=False)
    group.add_argument(
        '--group', action='store_true', dest='group',
        default=settings['group_repeating_emotes']['enabled'],
        help=('Group repeating emotes in all messages. '
              'Example: Kappa Kappa Kappa → Kappa x3.'))
    group.add_argument(
        '--no-group', action='store_false', dest='group',
        help='Opposite of --group.')

    settings_group.add_argument(
        '--group-threshold', metavar='repeats', type=int,
        default=settings['group_repeating_emotes']['threshold'],
        help='Minimal number of repeating emotes to group.')
    settings_group.add_argument(
        '--group-collocations', metavar='words', type=int,
        default=settings['group_repeating_emotes']['collocations'],
        help=('Maximum number of words in repeating collocations '
              '(default: 1, more is slower).'))
    settings_group.add_argument(
        '--group-collocations-threshold', metavar='repeats', type=int,
        default=settings['group_repeating_emotes']['collocations_threshold'],
        help=('Same logic as in --group-threshold, but applies only to '
              'repeating collocations.'))
    settings_group.add_argument(
        '--group-format', metavar='FORMAT', type=str,
        default=settings['group_repeating_emotes']['format'],
        help=('Python str.format for grouped emotes. Available '
              'variables: {emote} and {count}.'))

    badges = settings_group.add_mutually_exclusive_group(required=False)
    badges.add_argument(
        '--badges', action='store_true', dest='badges',
        default=settings['badges']['enabled'],
        help=('Add badges before usernames (IRC only).'))
    badges.add_argument(
        '--no-badges', action='store_false', dest='group',
        help='Opposite of --badges.')

    settings_group.add_argument(
        '--badges-max', metavar='n', type=int,
        default=settings['badges']['max_count'],
        help='Maximum number of badges per user.')

    channel_group = parser.add_argument_group(
        'Channel Mode Settings',
        'These options will only work with -c/--channel.')
    channel_group.add_argument(
        '--video-min', metavar='ID', type=int, default=0,
        help='ID of the earliest VOD to download.')
    channel_group.add_argument(
        '--video-max', metavar='ID', type=int, default=None,
        help='ID of the latest VOD to download.')
    channel_group.add_argument(
        '--video-count', metavar='N', type=int, default=None,
        help='Download N the most recent VODs.')
    channel_group.add_argument(
        '--video-types', metavar='type1,type2', type=str, default='archive',
        help=('Comma-separated list of VOD type to download. '
              'Available types: archive, upload, highlight, past_premiere.'))

    return parser


args, unknown_args = _post_init_parser(help=False).parse_known_args()
settings['client_id'] = args.client_id
settings['cooldown'] = args.cooldown
settings['display_progress'] = args.progress
settings['formats'] = args.formats
settings['directory'] = args.directory
settings['filename_format'] = args.filename_format
settings['max_width'] = args.max_width
settings['subtitle_duration'] = args.subtitle_duration
settings['dynamic_duration']['enabled'] = args.dynamic
settings['dynamic_duration']['max'] = args.dynamic_duration_max
settings['dynamic_duration']['max_length'] = args.dynamic_duration_max_length
settings['millisecond_separator'] = args.millisecond_separator
settings['group_repeating_emotes']['enabled'] = args.group
settings['group_repeating_emotes']['threshold'] = args.group_threshold
settings['group_repeating_emotes']['collocations'] = args.group_collocations
settings['group_repeating_emotes']['collocations_threshold'] = args.group_collocations_threshold
settings['group_repeating_emotes']['format'] = args.group_format
settings['badges']['enabled'] = args.badges
settings['badges']['max_count'] = args.badges_max
settings['video_types'] = args.video_types


#
# Post-init arguments
#

argparser = _post_init_parser(help=True)

source_group = argparser.add_argument_group('Chat Sources')
exclusive_group = source_group.add_mutually_exclusive_group(required=True)
exclusive_group.add_argument(
    '-c', '--channel', metavar='NAME', type=str,
    help=('Name of channel to download ALL chats from. '
          'Example: twitch.tv/{channel}.'))
exclusive_group.add_argument(
    '-v', '--video', metavar='ID', type=int,
    help='ID of VOD. Example: twitch.tv/videos/{video}.')
exclusive_group.add_argument(
    'video_id', type=int, metavar='VIDEO_ID', nargs='?',
    help='Alias for -v/--video (for backward compatibility).')
exclusive_group.add_argument(
    '--generate-config', action='store_true',
    help=('Generate settings.json in current directory using defaults and '
          'command-line arguments.'))



from time import sleep
from iso8601 import parse_date as parse8601
from progressbar import ProgressBar
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from .settings import settings


client = Session()
client.headers['Acccept'] = '*/*'
client.headers['Client-ID'] = settings['client_id']

# Configure retries for all requests
retries = Retry(connect=5, read=2, redirect=5)
http_adapter = HTTPAdapter(max_retries=retries)
client.mount('http://', http_adapter)
client.mount('https://', http_adapter)


def gql(query: str):
    res = client.post('https://gql.twitch.tv/gql', json={'query': query})

    if res.status_code == 200:
        return res.json()
    else:
        raise Exception(res.text)


class Message(object):
    @staticmethod
    def _find_groups(words, threshold=3, collocations=1,
                     collocations_threshold=2):
        groups = []
        words = words.copy()

        for size in range(min(collocations, len(words) // threshold + 1), 0, -1):
            for pos in range(len(words) - size):
                chunk = words[pos:pos+size]

                if None in chunk or \
                   len(Message._find_groups(chunk, threshold=2)) > 0:
                    continue

                count = 1
                for j in range(1, len(words) // size):
                    if chunk == words[pos+j*size:pos+j*size+size]:
                        count += 1
                    else:
                        break

                if count >= threshold or \
                   len(chunk) > 1 and count >= collocations_threshold:
                    groups.append((chunk, pos, count))
                    words[pos:pos+size*count] = [None] * size * count
        
        return groups

    @staticmethod
    def group(message, threshold=3, collocations=1, collocations_threshold=2,
              format='{emote} x{count}', **kwargs):
        words = message.split(' ')

        if len(words) < threshold:
            return message

        groups = Message._find_groups(words, threshold, collocations,
                                      collocations_threshold)
        groups = sorted(groups, key=lambda x: x[1], reverse=True)

        for chunk, pos, count in groups:
            emote = ' '.join(chunk)  # thin space!
            words = words[:pos] + \
                [format.format(emote=emote, count=count)] + \
                words[pos + count * len(chunk):]

        return ' '.join(words)

    def __init__(self, comment):
        self.user = comment['commenter']['displayName']

        if settings['badges']['enabled']:
            badge_ids = [badge['setID']
                         for badge in comment['commenter']['displayBadges']]

            badges = [val
                      for key, val in settings['badges']['map'].items()
                      if key in badge_ids]

            max_count = settings['badges']['max_count']
            if max_count >= 1:
                if len(badges) > max_count:
                    badges = badges[0:max_count]

            self.badge = ''.join(badges)
        else:
            self.badge = ''

        group_prefs = settings.get('group_repeating_emotes')

        message = ''.join(frag['text']
                          for frag in comment['message']['fragments']).strip()

        if group_prefs['enabled'] is True:
            self.message = self.group(message, **group_prefs)
        else:
            self.message = message

        self.offset = comment['contentOffsetSeconds']

        if comment['message']['userColor']:
            self.color = comment['message']['userColor'][1:]
        else:
            self.color = 'FFFFFF'

    def hash(self) -> int:
        return hash((self.offset, self.user, self.message))

class Messages(object):
    def __init__(self, video_id):
        self.video_id = video_id

        video = gql(f'''
            query {{
                video(id: {video_id}) {{
                    creator {{
                        displayName
                        id
                    }}
                    createdAt
                    lengthSeconds
                    title
                }}
            }}
        ''')

        self.created_at = parse8601(video['data']['video']['createdAt'])
        self.duration = video['data']['video']['lengthSeconds']
        self.title = video['data']['video']['title']
        self.creator_name = video['data']['video']['creator']['displayName']
        self.creator_id = video['data']['video']['creator']['id']

        if settings.get('display_progress') in [None, True]:
            self.progressbar = ProgressBar(max_value=self.duration)
    
    def __iter__(self):
        hasNextPage = True
        cursor = 0
        hashes = set()

        while hasNextPage:
            res = gql(f'''
                query {{
                    video(id: "{self.video_id}") {{
                        comments{f'(contentOffsetSeconds: {cursor})' if cursor > 0 else ''} {{
                            edges {{
                                cursor
                                node {{
                                    commenter {{
                                        displayName
                                        login
                                        displayBadges(channelID: {self.creator_id}) {{
                                            setID
                                        }}
                                    }}
                                    createdAt
                                    contentOffsetSeconds
                                    message {{
                                        fragments {{
                                            text
                                        }}
                                        userColor
                                    }}
                                }}
                            }}

                            pageInfo {{
                                hasNextPage
                            }}
                        }}
                    }}
                }}
            ''')

            comments = res['data']['video']['comments']
            hasNextPage = comments['pageInfo']['hasNextPage']

            # Avoid infinite loops
            new_cursor = comments['edges'][-1]['node']['contentOffsetSeconds']
            if new_cursor <= cursor:
                cursor += 1
                continue
            else:
                cursor = new_cursor

            for comment in comments['edges']:
                # Calculate more accurate offset
                ts = parse8601(comment['node']['createdAt'])
                offset = (ts - self.created_at).total_seconds()
                comment['node']['contentOffsetSeconds'] = offset

                try:
                    msg = Message(comment['node'])
                except Exception:
                    continue
            
                msg_hash = msg.hash()

                if msg_hash in hashes:
                    continue
                else:
                    hashes.add(msg_hash)

                yield msg

            if self.progressbar:
                ts = comments['edges'][-1]['node']['contentOffsetSeconds']
                self.progressbar.update(min(self.duration, ts))

            if settings['cooldown'] > 0:
                sleep(settings['cooldown'] / 1000)


class Channel(object):
    def __init__(self, channel):
        self.name = channel

    def videos(self):
        hasNextPage = True
        cursor = None

        while hasNextPage:
            res = gql(f'''
                query {{
                    user(login: "{self.name}") {{
                        videos({f'after: "{cursor}"' if cursor else 'type: ARCHIVE'}) {{
                            edges {{
                                cursor
                                node {{
                                    id
                                    createdAt
                                }}
                            }}
                            pageInfo {{
                                hasNextPage
                            }}
                        }}
                    }}
                }}
            ''')

            videos = res['data']['user']['videos']
            hasNextPage = videos['pageInfo']['hasNextPage']
            cursor = videos['edges'][-1]['cursor']

            for video in res['data']['user']['videos']['edges']:
                yield int(video['node']['id'])



from setuptools import setup, find_packages


with open('README.md', 'r') as fi:
    long_description = fi.read()


setup(
    name='tdh-tcd',
    version='2.5.2',

    author='Dmitry Karikh',
    author_email='the.dr.hax@gmail.com',

    description='Twitch Chat Downloader',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/TheDrHax/Twitch-Chat-Downloader',

    install_requires=[
        'requests',
        'progressbar2',
        'iso8601'
    ],

    packages=find_packages(),
    package_data={'tcd': ['example.settings.json']},
    include_package_data=True,

    entry_points='''
    [console_scripts]
    tcd=tcd:main
    ''',

    classifiers=[
        'Environment :: Console',
        'Programming Language :: Python :: 3.6',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Utilities'
    ]
)



from __future__ import unicode_literals

import io
import os
import re
import textwrap

from datetime import timedelta, datetime as dtt

from .settings import settings
from .twitch import Messages


class Subtitle(object):
    def __init__(self, filename: str):
        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))

        self.file = io.open(filename, mode='w+', encoding='UTF8')

    @staticmethod
    def _duration(msg):
        T_MIN = settings['subtitle_duration']
        T_MAX = settings['dynamic_duration']['max']
        MSG_MAX = settings['dynamic_duration']['max_length']

        if settings['dynamic_duration']['enabled']:
            part = (MSG_MAX - min(len(msg), MSG_MAX)) / MSG_MAX
            return T_MAX - (T_MAX - T_MIN) * part
        else:
            return T_MIN

    @staticmethod
    def ftime(seconds: float) -> str:
        t = dtt.strptime('00:00', '%H:%M') + timedelta(seconds=seconds)
        ts = dtt.strftime(t, '%H:%M:%S.%f')
        return ts[1:] if ts.startswith('00') else ts

    @staticmethod
    def wrap(username, text):
        max_width = settings['max_width']
        full_text = username + ': ' + text

        if len(full_text) <= max_width or max_width <= 0:
            return text

        text = textwrap.wrap(full_text, max_width, drop_whitespace=False)
        text = '\n'.join(text).replace('\n ', ' \n')
        text = text[len(username)+2:]

        return text

    def close(self):
        self.file.flush()
        self.file.close()


class SubtitlesASS(Subtitle):
    def __init__(self, filename: str):
        super(SubtitlesASS, self).__init__(filename)

        self.line = settings['ssa_events_line_format'] + '\n'

        self.file.writelines([
            '[Script Info]\n',
            'PlayResX: 1280\n',
            'PlayResY: 720\n',
            '\n',
            '[V4 Styles]\n',
            settings['ssa_style_format'],
            '\n',
            settings['ssa_style_default'],
            '\n\n',
            '[Events]\n',
            settings['ssa_events_format'],
            '\n'
        ])

    @staticmethod
    def _rgb_to_bgr(color):
        return color[4:6] + color[2:4] + color[0:2]

    @staticmethod
    def _color(text, color):
        return '{\\c&H' + color + '&}' + text + '{\\c&HFFFFFF&}'

    @staticmethod
    def wrap(username, message):
        return Subtitle.wrap(username, message).replace('\n', '\\N')

    @staticmethod
    def ftime(seconds: float) -> str:
        return Subtitle.ftime(seconds)[:-4]

    def add(self, comment):
        offset = round(comment.offset, 2)
        color = self._rgb_to_bgr(comment.color)

        self.file.write(self.line.format(
            start=self.ftime(offset),
            end=self.ftime(offset + self._duration(comment.message)),
            user=self._color(comment.user, color),
            message=self.wrap(comment.user, comment.message)
        ))


class SubtitlesSRT(Subtitle):
    def __init__(self, filename: str):
        super(SubtitlesSRT, self).__init__(filename)
        self.count = 0

    @staticmethod
    def ftime(seconds):
        return Subtitle.ftime(seconds)[:-3].replace('.', ',')

    def add(self, comment):
        time = comment.offset

        self.file.write(str(self.count) + '\n')
        self.file.write('{start} --> {end}\n'.format(
            start=self.ftime(time),
            end=self.ftime(time + self._duration(comment.message))
        ))
        self.file.write('{user}: {message}\n\n'.format(
            user=comment.user,
            message=comment.message
        ))

        self.count += 1


class SubtitlesIRC(Subtitle):
    def __init__(self, filename: str):
        super(SubtitlesIRC, self).__init__(filename)

    @staticmethod
    def ftime(seconds):
        return Subtitle.ftime(seconds)[:-3].replace('.', settings['millisecond_separator'])

    def add(self, comment):
        self.file.write('[{start}] <{badge}{user}> {message}\n'.format(
            start=self.ftime(comment.offset),
            badge=comment.badge,
            user=comment.user,
            message=comment.message
        ))


class SubtitleWriter:
    @staticmethod
    def clean_filename(string, valid_filename_regex=re.compile(r'(?u)[^-\w.()\[\]{}@%! ]')):
        return re.sub(valid_filename_regex, '', string.strip())

    @classmethod
    def filename(cls, video: Messages, ext: str) -> str:
        # TODO: make time configurable
        time_str = str(video.created_at.replace(microsecond=0).replace(tzinfo=None).isoformat())

        return settings['filename_format'].format(
            directory=settings['directory'],
            video_id=video.video_id,
            format=ext,
            user_name=cls.clean_filename(video.creator_name),
            title=cls.clean_filename(video.title),
            created_at=cls.clean_filename(time_str)
        )

    def __init__(self, video: Messages):
        self.drivers = set()

        for ext in settings['formats']:
            filename = self.filename(video, ext)

            if ext in ('ass', 'ssa'):
                self.drivers.add(SubtitlesASS(filename))

            if ext == 'srt':
                self.drivers.add(SubtitlesSRT(filename))

            if ext == 'irc':
                self.drivers.add(SubtitlesIRC(filename))

    def add(self, comment):
        [driver.add(comment) for driver in self.drivers]

    def close(self):
        [driver.close() for driver in self.drivers]
